/*****************************************************************************/
/*  bookmarks.c - routines for the bookmarks                                 */
/*  Copyright (C) 1998-1999 Brian Masney <masneyb@newwave.net>               */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either version 2 of the License, or        */
/*  (at your option) any later version.                                      */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA      */
/*****************************************************************************/

#include "ftp.h"

extern GtkWidget *hostedit, *portedit, *useredit, *passedit, *firewall_btn;
extern GdkPixmap *dir_pixmap, *open_dir_pixmap;
extern GdkBitmap *dir_mask, *open_dir_mask;
extern GHashTable *bookmarks_htable;
extern struct conn_categories *hostcat;
extern struct ftp_window_data window2;
extern GtkItemFactory *factory;
extern char emailaddr[MAXSTR];
extern int use_firewall;

static void doadd_bookmark (GtkWidget *widget, gpointer *data);
static void make_nonnull (char **str);
static void bm_apply_changes (GtkWidget *widget, gpointer backup_data);
static struct conn_categories *copy_bookmarks (struct conn_categories *bookmarks);
static void bm_close_dialog (GtkWidget *widget, GtkWidget *dialog);
static void after_move (GtkCTree *ctree, GtkCTreeNode *child, GtkCTreeNode *parent,
	GtkCTreeNode *sibling, gpointer data);
static void bm_dblclick (GtkWidget *widget, GdkEventButton *event, gpointer data);
static void bm_enter (GtkWidget *widget, GdkEventKey *event, gpointer data);
static gint hash_compare (gconstpointer path1, gconstpointer path2);
static guint hash_function (gconstpointer key);
static void new_folder_entry (gpointer data);
static void do_make_new (GtkWidget *widget, gpointer data);
static void new_item_entry (gpointer data);
static void edit_entry (gpointer data);
static void delete_entry (gpointer data);
static void entry_apply_changes (GtkWidget *widget, struct conn_categories *entry);
static void set_userpass_visible(GtkWidget *checkbutton, GtkWidget *entry);
void display_bm_struct (struct conn_categories *bm);

static GtkWidget *edit, *bm_hostedit, *bm_portedit, *bm_diredit, *bm_useredit, 
	*bm_passedit, *anon_chk, *password_chk, *firewall_check, *bm_pathedit, 
	*tree;
static struct conn_categories *new_bookmarks;
static GHashTable *htable, *new_bookmarks_htable;

void run_bookmark (gpointer data) {
   struct conn_categories *tempentry;
   
   tempentry = (struct conn_categories *) 
   	g_hash_table_lookup (bookmarks_htable, (char *) data);
   if (window2.local != -1) disconnect (&window2);
   gftp_set_username (window2.hdata->ftpdata, tempentry->user);
   if (strncmp (tempentry->pass, "@EMAIL@", 7) == 0) {
      gftp_set_password (window2.hdata->ftpdata, emailaddr);
   }
   else {
      gftp_set_password (window2.hdata->ftpdata, tempentry->pass);
   }
   gftp_set_hostname (window2.hdata->ftpdata, tempentry->hostname);
   gftp_set_directory (window2.hdata->ftpdata, tempentry->dir);
   gftp_set_port (window2.hdata->ftpdata, strtol(tempentry->port, (char **) NULL, 10));
   if (*tempentry->firewall != '1') {
      window2.hdata->ftpdata->proxy_type = 0;
   }
   else {
      window2.hdata->ftpdata->proxy_type = use_firewall;
   }
   ftp_connect (window2.hdata);
}
/*****************************************************************************/
void add_bookmark (gpointer data) {
   edit = MakeEditDialog("Add Bookmark", "Enter the name of the bookmark you want to add\nYou can separate items by a / to put it into a submenu\n(ex: Linux Sites/Debian)",
   	"Change", doadd_bookmark, data,
   	"  Cancel  ", NULL, NULL);
}
/*****************************************************************************/
static void doadd_bookmark (GtkWidget *widget, gpointer *data) {
   GtkItemFactoryEntry test = {NULL, NULL, run_bookmark, 0};
   struct conn_categories *tempentry;
   char *edttxt;
   
   if (hostcat->children == NULL) {
      hostcat->children = g_malloc0 (sizeof (struct conn_categories));
      tempentry = hostcat->children;
   }
   else {
      tempentry = hostcat->children;
      while (tempentry->next != NULL) tempentry = tempentry->next;
      tempentry->next = g_malloc0 (sizeof (struct conn_categories));
      tempentry = tempentry->next;
   }
   tempentry->prev = hostcat;

   edttxt = gtk_entry_get_text (GTK_ENTRY (edit));
   tempentry->path = g_malloc (strlen (edttxt) + 1);
   strcpy (tempentry->path, edttxt);
   make_nonnull (&tempentry->path);

   edttxt = gtk_entry_get_text (GTK_ENTRY (hostedit));
   tempentry->hostname = g_malloc (strlen (edttxt) + 1);
   strcpy (tempentry->hostname, edttxt);
   make_nonnull (&tempentry->hostname);

   edttxt = gtk_entry_get_text (GTK_ENTRY (portedit));
   tempentry->port = g_malloc (strlen (edttxt) + 1);
   strcpy (tempentry->port, edttxt);
   make_nonnull (&tempentry->port);

   edttxt = gtk_entry_get_text (GTK_ENTRY (window2.diredit));
   tempentry->dir = g_malloc (strlen (edttxt) + 1);
   strcpy (tempentry->dir, edttxt);
   make_nonnull (&tempentry->dir);

   edttxt = gtk_entry_get_text (GTK_ENTRY (useredit));
   tempentry->user = g_malloc (strlen (edttxt) + 1);
   strcpy (tempentry->user, edttxt);
   make_nonnull (&tempentry->user);

   edttxt = gtk_entry_get_text (GTK_ENTRY (passedit));
   tempentry->pass = g_malloc (strlen (edttxt) + 1);
   strcpy (tempentry->pass, edttxt);
   make_nonnull (&tempentry->pass);

   tempentry->firewall = g_malloc (1);
   *tempentry->firewall = GTK_CHECK_MENU_ITEM (firewall_btn)->active;

   g_hash_table_insert (bookmarks_htable, tempentry->path, tempentry);

   test.path = g_strjoin (NULL, "/Bookmarks/", tempentry->path, NULL);
   gtk_item_factory_create_item (factory, &test, (gpointer) tempentry->path, 1);
   g_free (test.path);
   write_config_file ();
}
/*****************************************************************************/
static void make_nonnull (char **str) {
   if (*str == NULL) {
      *str = g_malloc (1);
      **str = '\0';
   }
}
/*****************************************************************************/
void edit_bookmarks (gpointer data) { 
   GtkWidget *dialog, *scroll, *tempwid;
   GtkItemFactory *ifactory;
   struct conn_categories *tempentry, *preventry; 
   char *pos, tempchar, *prevpos, *text[2];
   GtkCTreeNode *parent;
   GtkItemFactoryEntry menu_items[] = {
      {"/_File",           	NULL,   0,      	0, 	"<Branch>"},
      {"/File/tearoff",  	NULL,   0,		0,	"<Tearoff>"},
      {"/File/New Folder...", 		NULL, 	new_folder_entry,0},
      {"/File/New Item...", 		NULL, 	new_item_entry,0},
      {"/File/Delete", 		NULL, 	delete_entry,	0},
      {"/File/Properties...", 	NULL, 	edit_entry,	0},
      {"/File/sep",  		NULL,   0,		0,	"<Separator>"},
      {"/File/Close",		NULL,	gtk_widget_destroy,	0}};
   
   htable = g_hash_table_new (hash_function, hash_compare);
   new_bookmarks = copy_bookmarks (hostcat);
   new_bookmarks_htable = build_hash_table (new_bookmarks);
   
   dialog = gtk_dialog_new ();
   gtk_window_set_title (GTK_WINDOW (dialog), "Edit Bookmarks");
   gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->action_area), 15);
   gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
   
   ifactory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>", NULL);
   gtk_item_factory_create_items (ifactory, 7, menu_items, NULL);
   gtk_item_factory_create_items (ifactory, 1, menu_items + 6, dialog);
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), ifactory->widget, FALSE, FALSE, FALSE);
   gtk_widget_show (ifactory->widget);
   
   scroll = gtk_scrolled_window_new (NULL, NULL);
   gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroll), 
   	GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
   gtk_widget_set_usize (scroll, 150, 200);
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), scroll, TRUE, TRUE, TRUE);
   gtk_container_border_width (GTK_CONTAINER (scroll), 3);
   gtk_widget_show (scroll);

   tree = gtk_ctree_new (1, 0);
   gtk_clist_set_selection_mode (GTK_CLIST (tree), GTK_SELECTION_SINGLE);
   gtk_clist_set_reorderable (GTK_CLIST (tree), 1);
   gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scroll), tree);
   gtk_signal_connect_after (GTK_OBJECT (tree), "button_press_event", GTK_SIGNAL_FUNC (bm_dblclick), (gpointer) tree);
   gtk_signal_connect_after (GTK_OBJECT (tree), "key_press_event", GTK_SIGNAL_FUNC (bm_enter), (gpointer) tree);
   gtk_signal_connect_after (GTK_OBJECT (tree), "tree_move", GTK_SIGNAL_FUNC (after_move), NULL);
   gtk_widget_show (tree);
   
   text[0] = text[1] = "Bookmarks";
   parent = gtk_ctree_insert_node (GTK_CTREE (tree), NULL, NULL, text, 5, 
   	dir_pixmap, dir_mask, open_dir_pixmap, open_dir_mask, FALSE, TRUE);
   new_bookmarks->cnode = parent;
   g_hash_table_insert (htable, new_bookmarks->cnode, new_bookmarks);
   
   tempentry = new_bookmarks->children;
   while (tempentry != NULL) {
      if (tempentry->children != NULL) {
         tempentry = tempentry->children;
         continue;
      }
      else {
         pos = tempentry->path;
         prevpos = pos;
         while ((pos = strchr (pos, '/')) != NULL) {
            tempchar = *pos;
            *pos = '\0';
            preventry = (struct conn_categories *) 
            	g_hash_table_lookup (new_bookmarks_htable, tempentry->path);
            if (preventry->cnode == NULL) {
               text[0] = text[1] = prevpos;
               preventry->cnode = gtk_ctree_insert_node (GTK_CTREE (tree), 
               		preventry->prev->cnode, NULL, text, 5, dir_pixmap, dir_mask, 
               		open_dir_pixmap, open_dir_mask, FALSE, FALSE);
               g_hash_table_insert (htable, preventry->cnode, preventry);
            }
            *pos = tempchar;
            prevpos = ++pos;
         }
      }

      if ((pos = strrchr (tempentry->path, '/')) == NULL) pos = tempentry->path;
      else pos++;
 
      text[0] = text[1] = pos;
      tempentry->cnode = gtk_ctree_insert_node (GTK_CTREE (tree), 
      		tempentry->prev->cnode, NULL, text, 5, NULL, NULL, 
      		NULL, NULL, FALSE, FALSE);
      g_hash_table_insert (htable, tempentry->cnode, tempentry);

      while (tempentry->next == NULL && tempentry->prev != NULL) tempentry = tempentry->prev;
      tempentry = tempentry->next;
   }

   tempwid = gtk_button_new_with_label ("OK");
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, FALSE);
   gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (bm_apply_changes), (gpointer) NULL);
   gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (bm_close_dialog), (gpointer) dialog);
   GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
   gtk_widget_show (tempwid);

   tempwid = gtk_button_new_with_label ("  Cancel  ");
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, FALSE);
   gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (bm_close_dialog), (gpointer) dialog);
   GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
   gtk_widget_grab_focus (tempwid);
   gtk_widget_show (tempwid);

   tempwid = gtk_button_new_with_label ("Apply");
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, FALSE);
   gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (bm_apply_changes), (gpointer) 1);
   GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
   gtk_widget_show (tempwid);

   gtk_widget_show (dialog);
}
/*****************************************************************************/
static void bm_apply_changes (GtkWidget *widget, gpointer backup_data) {
   struct conn_categories *tempentry, *delentry;
   char tempstr[MAXSTR];
   
   tempentry = hostcat->children;
   while (tempentry != NULL) {
      if (tempentry->path) {
         if (!tempentry->isfolder) {
            g_snprintf (tempstr, sizeof (tempstr), "/Bookmarks/%s", tempentry->path);
            tempstr[sizeof (tempstr) - 1] = '\0';

/* The gtk_item_factory_delete_item () has a bug in it. This will be fixed in
   version 1.2.4 of gtk+. I will uncomment these in my next version */

/*FIXME            gtk_item_factory_delete_item (factory, tempstr);*/
            gtk_widget_destroy (gtk_item_factory_get_item (factory, tempstr));
            g_free (tempentry->path);
         }
      }
      if (tempentry->hostname) g_free (tempentry->hostname);
      if (tempentry->dir) g_free (tempentry->dir);
      if (tempentry->user) g_free (tempentry->user);
      if (tempentry->pass) g_free (tempentry->pass);
      if (tempentry->port) g_free (tempentry->port);
      if (tempentry->firewall) g_free (tempentry->firewall);

      if (tempentry->children != NULL) {
         tempentry = tempentry->children;
      }
      else {
         if (tempentry->next == NULL) {
            while (tempentry->next == NULL && tempentry->prev != NULL) {
               delentry = tempentry;
               tempentry = tempentry->prev;
               if (delentry->isfolder) {
                  g_snprintf (tempstr, sizeof (tempstr), "/Bookmarks/%s", delentry->path);
                  tempstr[sizeof (tempstr) - 1] = '\0';
/*FIXME                  gtk_item_factory_delete_item (factory, tempstr);*/
                  gtk_widget_destroy (gtk_item_factory_get_item (factory, tempstr));
                  g_free (delentry->path);
               }
               g_free (delentry);
            }
            delentry = tempentry;
            tempentry = tempentry->next;
            if (delentry->isfolder && tempentry != NULL) {
               g_snprintf (tempstr, sizeof (tempstr), "/Bookmarks/%s", delentry->path);
               tempstr[sizeof (tempstr) - 1] = '\0';
/*FIXME               gtk_item_factory_delete_item (factory, tempstr);*/
               gtk_widget_destroy (gtk_item_factory_get_item (factory, tempstr));
               g_free (delentry->path);
               g_free (delentry);
            }
         }
         else {
            delentry = tempentry;
            tempentry = tempentry->next;
            g_free (delentry);
         }
      }
   }
   g_free (hostcat);
   g_hash_table_destroy (bookmarks_htable);
   hostcat = new_bookmarks;
   bookmarks_htable = new_bookmarks_htable;
   if (backup_data) {
      new_bookmarks = copy_bookmarks (hostcat);
      new_bookmarks_htable = build_hash_table (new_bookmarks);
   }
   else {
      new_bookmarks = NULL;
      new_bookmarks_htable = NULL;
   }
   build_bookmarks_menu ();
}
/*****************************************************************************/
static struct conn_categories *copy_bookmarks (struct conn_categories *bookmarks) {
   struct conn_categories *new_bm, *preventry, *tempentry, *sibling, 
      *newentry, *tentry;
   
   new_bm = g_malloc0 (sizeof (struct conn_categories));
   new_bm->path = g_malloc (1);
   *new_bm->path = '\0';
   new_bm->isfolder = bookmarks->isfolder;
   preventry = new_bm;
   tempentry = bookmarks->children;
   sibling = NULL;
   while (tempentry != NULL) {
      newentry = g_malloc0 (sizeof (struct conn_categories));
      newentry->isfolder = tempentry->isfolder;
      if (tempentry->path) {
         newentry->path = g_malloc (strlen (tempentry->path) + 1);
         strcpy (newentry->path, tempentry->path);
      }
      if (tempentry->hostname) {
         newentry->hostname = g_malloc (strlen (tempentry->hostname) + 1);
         strcpy (newentry->hostname, tempentry->hostname);
      }
      if (tempentry->dir) {
         newentry->dir = g_malloc (strlen (tempentry->dir) + 1);
         strcpy (newentry->dir, tempentry->dir);
      }
      if (tempentry->user) {
         newentry->user = g_malloc (strlen (tempentry->user) + 1);
         strcpy (newentry->user, tempentry->user);
      }
      if (tempentry->pass) {
         newentry->pass = g_malloc (strlen (tempentry->pass) + 1);
         strcpy (newentry->pass, tempentry->pass);
      }
      if (tempentry->port) {
         newentry->port = g_malloc (strlen (tempentry->port) + 1);
         strcpy (newentry->port, tempentry->port);
      }
      if (tempentry->firewall) {
         newentry->firewall = g_malloc (strlen (tempentry->firewall) + 1);
         strcpy (newentry->firewall, tempentry->firewall);
      }
      if (sibling == NULL) {
         if (preventry->children == NULL) preventry->children = newentry;
         else {
            tentry = preventry->children;
            while (tentry->next != NULL) tentry = tentry->next;
            tentry->next = newentry;
         }
      }
      else sibling->next = newentry;
      newentry->prev = preventry;

      if (tempentry->children != NULL) {
         preventry = newentry;
         sibling = NULL;
         tempentry = tempentry->children;
      }
      else {
         if (tempentry->next == NULL) {
            sibling = NULL;
            while (tempentry->next == NULL && tempentry->prev != NULL) {
               tempentry = tempentry->prev;
               preventry = preventry->prev;
            }
            tempentry = tempentry->next;
         }
         else {
            sibling = newentry;
            tempentry = tempentry->next;
         }
      }
   }
   return (new_bm);
}
/*****************************************************************************/
static void bm_close_dialog (GtkWidget *widget, GtkWidget *dialog) {
   struct conn_categories *tempentry;
   
   if (new_bookmarks_htable) g_hash_table_destroy (new_bookmarks_htable);
   tempentry = new_bookmarks;
   while (tempentry != NULL) {
      tempentry->cnode = NULL;
      if (tempentry->children != NULL) {
         tempentry = tempentry->children;
      }
      else {
         while (tempentry->next == NULL && tempentry->prev != NULL) tempentry = tempentry->prev;
         tempentry = tempentry->next;
      }
   }
   new_bookmarks = NULL;
   gtk_widget_destroy (dialog);
}
/*****************************************************************************/
static void after_move (GtkCTree *ctree, GtkCTreeNode *child, GtkCTreeNode *parent,
	GtkCTreeNode *sibling, gpointer data) {
  
   struct conn_categories *childentry, *siblingentry, *parententry, *tempentry;
   char tempstr[MAXSTR], *pos;
   
   childentry = (struct conn_categories *) g_hash_table_lookup (htable, child);
   parententry = (struct conn_categories *) g_hash_table_lookup (htable, parent);
   siblingentry = (struct conn_categories *) g_hash_table_lookup (htable, sibling);

   tempentry = childentry->prev->children;
   if (tempentry == childentry) {
      childentry->prev->children = childentry->prev->children->next;
   }
   else {
      while (tempentry->next != childentry) tempentry = tempentry->next;
      tempentry->next = childentry->next;
   }
   childentry->prev = parententry;
   if (!parententry->isfolder) {
      childentry->next = parententry->children;
      parententry->children = childentry;
      gtk_ctree_move (ctree, child, parententry->prev->cnode, parententry->cnode);
   }
   else {
      if (siblingentry == NULL) {
         childentry->next = parententry->children;
         parententry->children = childentry;
      }
      else {
         if (parententry->children == siblingentry) {
            childentry->next = parententry->children;
            parententry->children = childentry;
         }
         else {
            tempentry = parententry->children;
            while (tempentry->next != siblingentry) tempentry = tempentry->next;
            childentry->next = tempentry->next;
            tempentry->next = childentry;
         }
      }

      if ((pos = strchr (childentry->path, '/')) == NULL) pos = childentry->path;
      else pos++;
      g_snprintf (tempstr, sizeof (tempstr), "%s/%s", parententry->path, pos);
      g_free (childentry->path);
      childentry->path = g_malloc (strlen (tempstr) + 1);
      strcpy (childentry->path, tempstr);
      tempentry = childentry->children;
      while (tempentry != NULL) {
         if ((pos = strchr (tempentry->path, '/')) == NULL) pos = tempentry->path;
         else pos++;
         g_snprintf (tempstr, sizeof (tempstr), "%s/%s", tempentry->prev->path, pos);
         g_free (tempentry->path);
         tempentry->path = g_malloc (strlen (tempstr) + 1);
         strcpy (tempentry->path, tempstr);
         if (tempentry->children != NULL) tempentry = tempentry->children;
         else {
            if (tempentry->next == NULL) {
               if (tempentry->prev == childentry) {
                  tempentry = NULL;
               }
               else {
                  while (tempentry->next == NULL && tempentry->prev != NULL) tempentry = tempentry->prev;
                  tempentry = tempentry->next;
               }
            }
            else tempentry = tempentry->next;
         }
      }
   }
}
/*****************************************************************************/
static void bm_dblclick (GtkWidget *widget, GdkEventButton *event, gpointer data) {
   if (event->type == GDK_2BUTTON_PRESS) edit_entry (NULL);
}
/*****************************************************************************/
static void bm_enter (GtkWidget *widget, GdkEventKey *event, gpointer data) {
   if (event->type == GDK_KEY_PRESS) {
      if (event->keyval == GDK_Return) edit_entry (NULL);
      else if (event->keyval == GDK_KP_Delete || event->keyval == GDK_Delete) {
         delete_entry (NULL);
      }
   }
}
/*****************************************************************************/
static gint hash_compare (gconstpointer path1, gconstpointer path2) {
   return (path1 == path2);
}
/*****************************************************************************/
static guint hash_function (gconstpointer key) {
   return ((guint) key * 33);
}
/*****************************************************************************/
static void new_folder_entry (gpointer data) {
   edit = MakeEditDialog ("New Folder", "Enter the name of the new folder to create",
   	"Create", do_make_new, (void *) 1,
   	"  Cancel  ", NULL, NULL);
}
/*****************************************************************************/
static void new_item_entry (gpointer data) {
   edit = MakeEditDialog ("New Folder", "Enter the name of the new item to create",
   	"Create", do_make_new, 0,
   	"  Cancel  ", NULL, NULL);
}
/*****************************************************************************/
static void do_make_new (GtkWidget *widget, gpointer data) {
   struct conn_categories *tempentry, *newentry;
   GtkCTreeNode *sibling;
   char *str, *pos, *text[2];
   
   str = gtk_entry_get_text (GTK_ENTRY (edit));
   pos = str;
   while ((pos = strchr (pos, '/')) != NULL) *pos++ = ' ';

   newentry = g_malloc0 (sizeof (struct conn_categories));
   newentry->path = g_malloc (strlen (str) + 1);
   strcpy (newentry->path, str);
   newentry->prev = new_bookmarks;
   if (data) newentry->isfolder = 1;
   
   if (new_bookmarks->children == NULL) {
      new_bookmarks->children = newentry;
      sibling = NULL;
   }
   else {
      tempentry = new_bookmarks->children;
      while (tempentry->next != NULL) tempentry = tempentry->next;
      tempentry->next = newentry;
      sibling = tempentry->cnode;
   }

   text[0] = text[1] = str;
   if (newentry->isfolder) {
      newentry->cnode = gtk_ctree_insert_node (GTK_CTREE (tree), 
      		new_bookmarks->cnode, NULL, text, 5, dir_pixmap, dir_mask, 
      		open_dir_pixmap, open_dir_mask, FALSE, FALSE);
   }
   else {
      newentry->cnode = gtk_ctree_insert_node (GTK_CTREE (tree), 
   		new_bookmarks->cnode, NULL, text, 5, NULL, NULL, 
   		NULL, NULL, FALSE, FALSE);
   }
   g_hash_table_insert (htable, newentry->cnode, newentry);
   g_hash_table_insert (new_bookmarks_htable, newentry->path, newentry);
}
/*****************************************************************************/
static void edit_entry (gpointer data) {
   GtkWidget *table, *tempwid, *dialog;
   struct conn_categories *entry;
   char *pos;
   
   if (GTK_CLIST (tree)->selection == NULL) return;
   entry = (struct conn_categories *) g_hash_table_lookup (htable, GTK_CLIST (tree)->selection->data);
   if (entry == NULL) return;

   dialog = gtk_dialog_new ();
   gtk_grab_add (dialog);
   gtk_window_set_title (GTK_WINDOW (dialog), "Edit Entry");
   gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->action_area), 15);
   gtk_container_border_width (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), 10);
   gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
   
   tempwid = gtk_frame_new (NULL);
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), tempwid, TRUE, TRUE, FALSE);
   gtk_widget_show (tempwid);
   
   table = gtk_table_new (10, 2, FALSE);
   gtk_container_border_width (GTK_CONTAINER (table), 5);
   gtk_table_set_row_spacings (GTK_TABLE (table), 5);
   gtk_table_set_col_spacings (GTK_TABLE (table), 5);
   gtk_container_add (GTK_CONTAINER (tempwid), table);
   gtk_widget_show (table);
   
   tempwid = gtk_label_new ("Description:");
   gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 0, 1);
   gtk_widget_show (tempwid);

   bm_pathedit = gtk_entry_new_with_max_length (MAXSTR);
   gtk_table_attach_defaults (GTK_TABLE (table), bm_pathedit, 1, 2, 0, 1);
   if ((pos = strrchr (entry->path, '/')) == NULL) pos = entry->path;
   else pos++;
   if (pos) gtk_entry_set_text (GTK_ENTRY (bm_pathedit), pos);
   gtk_widget_show (bm_pathedit);

   tempwid = gtk_label_new ("Hostname:");
   gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 1, 2);
   gtk_widget_show (tempwid);

   bm_hostedit = gtk_entry_new_with_max_length (MAXSTR);
   gtk_table_attach_defaults (GTK_TABLE (table), bm_hostedit, 1, 2, 1, 2);
   if (entry->hostname) gtk_entry_set_text (GTK_ENTRY (bm_hostedit), entry->hostname);
   gtk_widget_set_sensitive (bm_hostedit, !entry->isfolder);
   gtk_widget_show (bm_hostedit);

   tempwid = gtk_label_new ("Port:");
   gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 2, 3);
   gtk_widget_show (tempwid);

   bm_portedit = gtk_entry_new_with_max_length (MAXSTR);
   gtk_table_attach_defaults (GTK_TABLE (table), bm_portedit, 1, 2, 2, 3);
   if (entry->port) gtk_entry_set_text (GTK_ENTRY (bm_portedit), entry->port);
   gtk_widget_set_sensitive (bm_portedit, !entry->isfolder);
   gtk_widget_show (bm_portedit);

   tempwid = gtk_label_new ("Directory:");
   gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 3, 4);
   gtk_widget_show (tempwid);

   bm_diredit = gtk_entry_new_with_max_length (MAXSTR);
   gtk_table_attach_defaults (GTK_TABLE (table), bm_diredit, 1, 2, 3, 4);
   if (entry->dir) gtk_entry_set_text (GTK_ENTRY (bm_diredit), entry->dir);
   gtk_widget_set_sensitive (bm_diredit, !entry->isfolder);
   gtk_widget_show (bm_diredit);

   tempwid = gtk_hseparator_new ();
   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 2, 4, 5);
   gtk_widget_show (tempwid);

   tempwid = gtk_label_new ("Username:");
   gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 5, 6);
   gtk_widget_show (tempwid);

   bm_useredit = gtk_entry_new_with_max_length (MAXSTR);
   gtk_table_attach_defaults (GTK_TABLE (table), bm_useredit, 1, 2, 5, 6);
   if (entry->user) gtk_entry_set_text (GTK_ENTRY (bm_useredit), entry->user);
   gtk_widget_set_sensitive (bm_useredit, !entry->isfolder);
   gtk_widget_show (bm_useredit);

   tempwid = gtk_label_new ("Password:");
   gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 6, 7);
   gtk_widget_show (tempwid);

   bm_passedit = gtk_entry_new_with_max_length (MAXSTR);
   gtk_table_attach_defaults (GTK_TABLE (table), bm_passedit, 1, 2, 6, 7);
   gtk_entry_set_visibility (GTK_ENTRY (bm_passedit), FALSE);
   if (entry->pass) gtk_entry_set_text (GTK_ENTRY (bm_passedit), entry->pass);
   gtk_widget_set_sensitive (bm_passedit, !entry->isfolder);
   gtk_widget_show (bm_passedit);

   anon_chk = gtk_check_button_new_with_label ("Log in as ANONYMOUS");
   gtk_table_attach_defaults (GTK_TABLE (table), anon_chk, 0, 2, 7, 8);
   gtk_signal_connect (GTK_OBJECT (anon_chk), "toggled", GTK_SIGNAL_FUNC (set_userpass_visible), NULL);
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (anon_chk), entry->user && strcmp (entry->user, ANON_LOGIN) == 0);
   gtk_widget_set_sensitive (anon_chk, !entry->isfolder);
   gtk_widget_show (anon_chk);

   password_chk = gtk_check_button_new_with_label ("Save Password");
   gtk_table_attach_defaults (GTK_TABLE (table), password_chk, 0, 2, 8, 9);
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (password_chk), entry->pass && *entry->pass != '\0');
   gtk_widget_set_sensitive (password_chk, !entry->isfolder);
   gtk_widget_show (password_chk);

   firewall_check = gtk_check_button_new_with_label ("Connect through FTP Proxy Server");
   gtk_table_attach_defaults (GTK_TABLE (table), firewall_check, 0, 2, 9, 10);
   if (!entry->isfolder) {
      gtk_widget_set_sensitive (firewall_check, 0);
   }
   else {
      if (use_firewall) gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (firewall_check), entry->firewall && entry->firewall[0] == '1');
      else gtk_widget_set_sensitive (firewall_check, 0);
   }
   gtk_widget_show (firewall_check);
   
   tempwid = gtk_button_new_with_label ("OK");
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, FALSE);
   gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (entry_apply_changes), (gpointer) entry);
   gtk_signal_connect_object (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dialog));
   GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
   gtk_widget_show (tempwid);

   tempwid = gtk_button_new_with_label ("  Cancel  ");
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, FALSE);
   gtk_signal_connect_object (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dialog));
   GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
   gtk_widget_grab_focus (tempwid);
   gtk_widget_show (tempwid);

   tempwid = gtk_button_new_with_label ("Apply");
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, FALSE);
   gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (entry_apply_changes), (gpointer) entry);
   GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
   gtk_widget_show (tempwid);
   
   gtk_widget_show (dialog);
}
/*****************************************************************************/
static void delete_entry (gpointer data) {
   struct conn_categories *entry, *tempentry, *delentry;
   
   if (GTK_CLIST (tree)->selection == NULL) return;
   entry = (struct conn_categories *) g_hash_table_lookup (htable, GTK_CLIST (tree)->selection->data);
   if (entry == NULL || entry->prev == NULL) return;

   g_hash_table_remove (htable, entry->cnode);
   g_hash_table_remove (new_bookmarks_htable, entry->path);
   gtk_ctree_remove_node (GTK_CTREE (tree), entry->cnode);
   if (entry->prev->children == entry) {
      entry->prev->children = entry->prev->children->next;
   }
   else {
      tempentry = entry->prev->children;
      while (tempentry->next != entry) tempentry = tempentry->next;
      tempentry->next = entry->next;
   }

   entry->prev = NULL;
   entry->next = NULL;
   tempentry = entry;
   while (tempentry != NULL) {
      if (tempentry->path) g_free (tempentry->path);
      if (tempentry->hostname) g_free (tempentry->hostname);
      if (tempentry->dir) g_free (tempentry->dir);
      if (tempentry->user) g_free (tempentry->user);
      if (tempentry->pass) g_free (tempentry->pass);
      if (tempentry->port) g_free (tempentry->port);
      if (tempentry->firewall) g_free (tempentry->firewall);

      if (tempentry->children != NULL) {
         tempentry = tempentry->children;
      }
      else if (tempentry->next == NULL && tempentry->prev != NULL) {
         delentry = tempentry->prev;
         g_free (tempentry);
         tempentry = delentry->next;
         if (delentry != entry) g_free (delentry);
      }
      else if (tempentry->next != NULL) tempentry = tempentry->next;
      else tempentry = NULL;
   }
   g_free (entry);
}
/*****************************************************************************/
static void entry_apply_changes (GtkWidget *widget, struct conn_categories *entry) {
   char *str, *pos;
   
   g_hash_table_remove (new_bookmarks_htable, entry->path);
   str = gtk_entry_get_text (GTK_ENTRY (bm_pathedit));
   if ((pos = strrchr (entry->path, '/')) == NULL) *entry->path = '\0';
   else *(pos+1) = '\0';
   pos = entry->path;
   entry->path = g_strjoin (NULL, entry->path, str, NULL);
   g_free (pos);
   g_hash_table_insert (new_bookmarks_htable, entry->path, entry);

   str = gtk_entry_get_text (GTK_ENTRY (bm_hostedit));
   if (entry->hostname) g_free (entry->hostname);
   entry->hostname = g_malloc (strlen (str) + 1);
   strcpy (entry->hostname, str);
   
   str = gtk_entry_get_text (GTK_ENTRY (bm_portedit));
   if (entry->port) g_free (entry->port);
   entry->port = g_malloc (strlen (str) + 1);
   strcpy (entry->port, str);
   
   str = gtk_entry_get_text (GTK_ENTRY (bm_diredit));
   if (entry->dir) g_free (entry->dir);
   entry->dir = g_malloc (strlen (str) + 1);
   strcpy (entry->dir, str);
   
   if (GTK_TOGGLE_BUTTON (anon_chk)->active) str = ANON_LOGIN;
   else str = gtk_entry_get_text (GTK_ENTRY (bm_useredit));
   if (entry->user) g_free (entry->user);
   entry->user = g_malloc (strlen (str) + 1);
   strcpy (entry->user, str);
   
   if (GTK_TOGGLE_BUTTON (anon_chk)->active) str = emailaddr;
   else {
      if (GTK_TOGGLE_BUTTON (password_chk)->active) {
         str = gtk_entry_get_text (GTK_ENTRY (bm_passedit));
      }
      else str = "";
   }
   if (entry->pass) g_free (entry->pass);
   entry->pass = g_malloc (strlen (str) + 1);
   strcpy (entry->pass, str);
   
   *entry->firewall = GTK_TOGGLE_BUTTON (firewall_check)->active ? '1' : '0';
}
/*****************************************************************************/
static void set_userpass_visible(GtkWidget *checkbutton, GtkWidget *entry) {
   gtk_widget_set_sensitive (bm_useredit, !GTK_TOGGLE_BUTTON (anon_chk)->active);
   gtk_widget_set_sensitive (bm_passedit, !GTK_TOGGLE_BUTTON (anon_chk)->active);
}
/*****************************************************************************/
void build_bookmarks_menu (void) {
   GtkItemFactoryEntry test = {NULL, NULL, run_bookmark, 0};
   struct conn_categories *tempentry;
   
   tempentry = hostcat->children;
   while (tempentry != NULL) {
      test.path = g_strjoin (NULL, "/Bookmarks/", tempentry->path, NULL);
      if (tempentry->isfolder) {
         test.item_type = "<Branch>";
         test.callback = NULL;
      }
      else {
         test.item_type = "";
         test.callback = run_bookmark;
      }
      gtk_item_factory_create_item (factory, &test, (gpointer) tempentry->path, 1);
      g_free (test.path);
      if (tempentry->children != NULL) {
         tempentry = tempentry->children;
         continue;
      }
      while (tempentry->next == NULL && tempentry->prev != NULL) tempentry = tempentry->prev;
      tempentry = tempentry->next;
   }
}   
/*****************************************************************************/
void display_bm_struct (struct conn_categories *bm) {
   struct conn_categories *tempentry;
   
   tempentry = bm->children;
   while (tempentry != NULL) {
      printf ("%s (%d)\n", tempentry->path, tempentry->isfolder);
      if (tempentry->children != NULL) {
         printf ("Going up\n");
         tempentry = tempentry->children;
      }
      else {
         while (tempentry->next == NULL && tempentry->prev != NULL) {
            printf ("Going down\n");
            tempentry = tempentry->prev;
         }
         tempentry = tempentry->next;
      }
   }
}
/*****************************************************************************/
